@@ -154,7 +154,7 @@ xdirname(const char *path)
 	if (!(p = strdup(path)))
 		err(1, "strdup");
 	if (!(b = dirname(p)))
-		err(1, "basename");
+		err(1, "dirname");
 	if (!(b = strdup(b)))
 		err(1, "strdup");
 	free(p);
@@ -188,7 +188,8 @@ mkdirp(const char *path)
 {
 	char tmp[PATH_MAX], *p;
 
-	strlcpy(tmp, path, sizeof(tmp));
+	if (strlcpy(tmp, path, sizeof(tmp)) >= sizeof(tmp))
+		errx(1, "path truncated: '%s'", path);
 	for (p = tmp + (tmp[0] == '/'); *p; p++) {
 		if (*p != '/')
 			continue;
@@ -234,12 +235,15 @@ printtimeshort(FILE *fp, const git_time *intime)
 }
 
 int
-writeheader(FILE *fp)
+writeheader(FILE *fp, const char *title)
 {
-	fputs("<!DOCTYPE HTML>"
+	fputs("<!DOCTYPE html>\n"
 		"<html dir=\"ltr\" lang=\"en\">\n<head>\n"
 		"<meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\" />\n"
 		"<meta http-equiv=\"Content-Language\" content=\"en\" />\n<title>", fp);
+	xmlencode(fp, title, strlen(title));
+	if (title[0] && stripped_name[0])
+		fputs(" - ", fp);
 	xmlencode(fp, stripped_name, strlen(stripped_name));
 	if (description[0])
 		fputs(" - ", fp);
@@ -266,7 +270,7 @@ writeheader(FILE *fp)
 	fputs("<tr><td></td><td>\n", fp);
 	fprintf(fp, "<a href=\"%slog.html\">Log</a> | ", relpath);
 	fprintf(fp, "<a href=\"%sfiles.html\">Files</a> | ", relpath);
-	fprintf(fp, "<a href=\"%srefs.html\">Refs/branches</a>", relpath);
+	fprintf(fp, "<a href=\"%srefs.html\">Refs</a>", relpath);
 	if (hasreadme)
 		fprintf(fp, " | <a href=\"%sfile/README.html\">README</a>", relpath);
 	if (haslicense)
@@ -282,11 +286,11 @@ writefooter(FILE *fp)
 	return !fputs("</div>\n</body>\n</html>\n", fp);
 }
 
-void
+int
 writeblobhtml(FILE *fp, const git_blob *blob)
 {
 	off_t i;
-	size_t n = 1;
+	size_t n = 0;
 	char *nfmt = "<a href=\"#l%d\" id=\"l%d\">%d</a>\n";
 	const char *s = git_blob_rawcontent(blob);
 	git_off_t len = git_blob_rawsize(blob);
@@ -294,6 +298,7 @@ writeblobhtml(FILE *fp, const git_blob *blob)
 	fputs("<table id=\"blob\"><tr><td class=\"num\"><pre>\n", fp);
 
 	if (len) {
+		n++;
 		fprintf(fp, nfmt, n, n, n);
 		for (i = 0; i < len - 1; i++) {
 			if (s[i] == '\n') {
@@ -306,6 +311,8 @@ writeblobhtml(FILE *fp, const git_blob *blob)
 	fputs("</pre></td><td><pre>\n", fp);
 	xmlencode(fp, s, (size_t)len);
 	fputs("</pre></td></tr></table>\n", fp);
+
+	return n;
 }
 
 void
@@ -329,16 +336,15 @@ printcommit(FILE *fp, struct commitinfo *ci)
 		printtime(fp, &(ci->author->when));
 		fputc('\n', fp);
 	}
-	fputc('\n', fp);
-
-	if (ci->msg)
+	if (ci->msg) {
+		fputc('\n', fp);
 		xmlencode(fp, ci->msg, strlen(ci->msg));
-
-	fputc('\n', fp);
+		fputc('\n', fp);
+	}
 }
 
 void
-printshowfile(struct commitinfo *ci)
+printshowfile(FILE *fp, struct commitinfo *ci)
 {
 	const git_diff_delta *delta;
 	const git_diff_hunk *hunk;
@@ -346,30 +352,19 @@ printshowfile(struct commitinfo *ci)
 	git_patch *patch;
 	git_buf statsbuf;
 	size_t ndeltas, nhunks, nhunklines;
-	FILE *fp;
 	size_t i, j, k;
-	char path[PATH_MAX];
 
-	snprintf(path, sizeof(path), "commit/%s.html", ci->oid);
-	/* check if file exists if so skip it */
-	if (!access(path, F_OK))
-		return;
-
-	fp = efopen(path, "w");
-	writeheader(fp);
-	fputs("<pre>\n", fp);
 	printcommit(fp, ci);
 
 	memset(&statsbuf, 0, sizeof(statsbuf));
 
 	/* diff stat */
-	if (ci->stats) {
-		if (!git_diff_stats_to_buf(&statsbuf, ci->stats,
-		    GIT_DIFF_STATS_FULL | GIT_DIFF_STATS_SHORT, 80)) {
-			if (statsbuf.ptr && statsbuf.ptr[0]) {
-				fputs("<b>Diffstat:</b>\n", fp);
-				fputs(statsbuf.ptr, fp);
-			}
+	if (ci->stats &&
+	    !git_diff_stats_to_buf(&statsbuf, ci->stats,
+                                   GIT_DIFF_STATS_FULL | GIT_DIFF_STATS_SHORT, 80)) {
+		if (statsbuf.ptr && statsbuf.ptr[0]) {
+			fputs("<b>Diffstat:</b>\n", fp);
+			xmlencode(fp, statsbuf.ptr, strlen(statsbuf.ptr));
 		}
 	}
 
@@ -423,35 +418,27 @@ printshowfile(struct commitinfo *ci)
 	}
 	git_buf_free(&statsbuf);
 
-	fputs("</pre>\n", fp);
-	writefooter(fp);
-	fclose(fp);
 	return;
 }
 
 int
-writelog(FILE *fp, const char *branch)
+writelog(FILE *fp, const git_oid *oid)
 {
 	struct commitinfo *ci;
-	const git_oid *oid;
 	git_revwalk *w = NULL;
-	git_object *obj = NULL;
 	git_oid id;
 	size_t len;
-
-	mkdir("commit", 0755);
-
-	if (git_revparse_single(&obj, repo, branch))
-		return -1;
-	oid = git_object_id(obj);
+	char path[PATH_MAX];
+	FILE *fpfile;
+	int r;
 
 	git_revwalk_new(&w, repo);
 	git_revwalk_push(w, oid);
 	git_revwalk_sorting(w, GIT_SORT_TIME);
 	git_revwalk_simplify_first_parent(w);
 
-	fputs("<table id=\"log\"><thead>\n<tr><td>Age</td><td>Commit message</td>"
-		  "<td>Author</td><td>Files</td><td class=\"num\">+</td>"
+	fputs("<table id=\"log\"><thead>\n<tr><td>Date</td><td>Commit message</td>"
+		  "<td>Author</td><td class=\"num\">Files</td><td class=\"num\">+</td>"
 		  "<td class=\"num\">-</td></tr>\n</thead><tbody>\n", fp);
 
 	while (!git_revwalk_next(&id, w)) {
@@ -486,14 +473,26 @@ writelog(FILE *fp, const char *branch)
 		fputs("</td></tr>\n", fp);
 
 		relpath = "../";
-		printshowfile(ci);
 
+		r = snprintf(path, sizeof(path), "commit/%s.html", ci->oid);
+		if (r == -1 || (size_t)r >= sizeof(path))
+			errx(1, "path truncated: 'commit/%s.html'", ci->oid);
+
+		/* check if file exists if so skip it */
+		if (access(path, F_OK)) {
+			fpfile = efopen(path, "w");
+			writeheader(fpfile, ci->summary);
+			fputs("<pre>", fpfile);
+			printshowfile(fpfile, ci);
+			fputs("</pre>\n", fpfile);
+			writefooter(fpfile);
+			fclose(fpfile);
+		}
 		commitinfo_free(ci);
 	}
 	fputs("</tbody></table>", fp);
 
 	git_revwalk_free(w);
-	git_object_free(obj);
 
 	relpath = "";
 
@@ -516,6 +515,16 @@ printcommitatom(FILE *fp, struct commitinfo *ci)
 		xmlencode(fp, ci->summary, strlen(ci->summary));
 		fputs("</title>\n", fp);
 	}
+	fprintf(fp, "<link rel=\"alternate\" type=\"text/html\" href=\"commit/%s.html\" />",
+	        ci->oid);
+
+	if (ci->author) {
+		fputs("<author><name>", fp);
+		xmlencode(fp, ci->author->name, strlen(ci->author->name));
+		fputs("</name>\n<email>", fp);
+		xmlencode(fp, ci->author->email, strlen(ci->author->email));
+		fputs("</email>\n</author>\n", fp);
+	}
 
 	fputs("<content type=\"text\">", fp);
 	fprintf(fp, "commit %s\n", ci->oid);
@@ -535,13 +544,7 @@ printcommitatom(FILE *fp, struct commitinfo *ci)
 		xmlencode(fp, ci->msg, strlen(ci->msg));
 	}
 	fputs("\n</content>\n", fp);
-	if (ci->author) {
-		fputs("<author><name>", fp);
-		xmlencode(fp, ci->author->name, strlen(ci->author->name));
-		fputs("</name>\n<email>", fp);
-		xmlencode(fp, ci->author->email, strlen(ci->author->email));
-		fputs("</email>\n</author>\n", fp);
-	}
+
 	fputs("</entry>\n", fp);
 }
 
@@ -579,40 +582,40 @@ writeatom(FILE *fp)
 }
 
 int
-writeblob(git_object *obj, const char *filename, git_off_t filesize)
+writeblob(git_object *obj, const char *fpath, const char *filename, git_off_t filesize)
 {
-	char fpath[PATH_MAX];
 	char tmp[PATH_MAX] = "";
-	char *d, *p;
+	char *d;
+	const char *p;
+	int lc = 0;
 	FILE *fp;
 
-	snprintf(fpath, sizeof(fpath), "file/%s.html", filename);
 	d = xdirname(fpath);
 	if (mkdirp(d)) {
 		free(d);
-		return 1;
+		return -1;
 	}
 	free(d);
 
 	p = fpath;
 	while (*p) {
-		if (*p == '/')
-			strlcat(tmp, "../", sizeof(tmp));
+		if (*p == '/' && strlcat(tmp, "../", sizeof(tmp)) >= sizeof(tmp))
+			errx(1, "path truncated: '../%s'", tmp);
 		p++;
 	}
 	relpath = tmp;
 
 	fp = efopen(fpath, "w");
-	writeheader(fp);
+	writeheader(fp, filename);
 	fputs("<p> ", fp);
 	xmlencode(fp, filename, strlen(filename));
-	fprintf(fp, " (%jub)", (uintmax_t)filesize);
+	fprintf(fp, " (%juB)", (uintmax_t)filesize);
 	fputs("</p><hr/>", fp);
 
 	if (git_blob_is_binary((git_blob *)obj)) {
 		fputs("<p>Binary file</p>\n", fp);
 	} else {
-		writeblobhtml(fp, (git_blob *)obj);
+		lc = writeblobhtml(fp, (git_blob *)obj);
 		if (ferror(fp))
 			err(1, "fwrite");
 	}
@@ -621,7 +624,7 @@ writeblob(git_object *obj, const char *filename, git_off_t filesize)
 
 	relpath = "";
 
-	return 0;
+	return lc;
 }
 
 const char *
@@ -670,26 +673,31 @@ int
 writefilestree(FILE *fp, git_tree *tree, const char *branch, const char *path)
 {
 	const git_tree_entry *entry = NULL;
-	const char *filename;
-	char filepath[PATH_MAX];
+	const char *entryname;
+	char filepath[PATH_MAX], entrypath[PATH_MAX];
 	git_object *obj = NULL;
 	git_off_t filesize;
 	size_t count, i;
-	int ret;
+	int lc, r, ret;
 
 	count = git_tree_entrycount(tree);
 	for (i = 0; i < count; i++) {
 		if (!(entry = git_tree_entry_byindex(tree, i)) ||
 		    git_tree_entry_to_object(&obj, repo, entry))
 			return -1;
-		filename = git_tree_entry_name(entry);
+		entryname = git_tree_entry_name(entry);
+		r = snprintf(entrypath, sizeof(entrypath), "%s%s%s",
+			 path, path[0] ? "/" : "", entryname);
+		if (r == -1 || (size_t)r >= sizeof(entrypath))
+			errx(1, "path truncated: '%s%s%s'",
+			        path, path[0] ? "/" : "", entryname);
 		switch (git_object_type(obj)) {
 		case GIT_OBJ_BLOB:
 			break;
 		case GIT_OBJ_TREE:
 			/* NOTE: recurses */
 			ret = writefilestree(fp, (git_tree *)obj, branch,
-			                     filename);
+			                     entrypath);
 			git_object_free(obj);
 			if (ret)
 				return ret;
@@ -698,35 +706,36 @@ writefilestree(FILE *fp, git_tree *tree, const char *branch, const char *path)
 			git_object_free(obj);
 			continue;
 		}
-		if (path[0]) {
-			snprintf(filepath, sizeof(filepath), "%s/%s",
-			         path, filename);
-			filename = filepath;
-		}
+
+		r = snprintf(filepath, sizeof(filepath), "file/%s%s%s.html",
+		         path, path[0] ? "/" : "", entryname);
+		if (r == -1 || (size_t)r >= sizeof(filepath))
+			errx(1, "path truncated: 'file/%s%s%s.html'",
+			        path, path[0] ? "/" : "", entryname);
+
 		filesize = git_blob_rawsize((git_blob *)obj);
 
-		fputs("<tr><td>", fp);
-		fprintf(fp, "%s", filemode(git_tree_entry_filemode(entry)));
-		fprintf(fp, "</td><td><a href=\"%sfile/", relpath);
-		xmlencode(fp, filename, strlen(filename));
-		fputs(".html\">", fp);
-		xmlencode(fp, filename, strlen(filename));
-		fputs("</a></td><td class=\"num\">", fp);
-		fprintf(fp, "%ju", (uintmax_t)filesize);
-		fputs("</td></tr>\n", fp);
+		lc = writeblob(obj, filepath, entryname, filesize);
 
-		writeblob(obj, filename, filesize);
+		fputs("<tr><td>", fp);
+		fputs(filemode(git_tree_entry_filemode(entry)), fp);
+		fprintf(fp, "</td><td><a href=\"%s%s\">", relpath, filepath);
+		xmlencode(fp, entrypath, strlen(entrypath));
+		fputs("</a></td><td class=\"num\">", fp);
+		if (showlinecount && lc > 0)
+			fprintf(fp, "%dL", lc);
+		else
+			fprintf(fp, "%juB", (uintmax_t)filesize);
+		fputs("</td></tr>\n", fp);
 	}
 
 	return 0;
 }
 
 int
-writefiles(FILE *fp, const char *branch)
+writefiles(FILE *fp, const git_oid *id, const char *branch)
 {
-	const git_oid *id;
 	git_tree *tree = NULL;
-	git_object *obj = NULL;
 	git_commit *commit = NULL;
 	int ret = -1;
 
@@ -734,9 +743,6 @@ writefiles(FILE *fp, const char *branch)
 	      "<td>Mode</td><td>Name</td><td class=\"num\">Size</td>"
 	      "</tr>\n</thead><tbody>\n", fp);
 
-	if (git_revparse_single(&obj, repo, branch))
-		goto err;
-	id = git_object_id(obj);
 	if (git_commit_lookup(&commit, repo, id) ||
 	    git_commit_tree(&tree, commit))
 		goto err;
@@ -745,7 +751,6 @@ writefiles(FILE *fp, const char *branch)
 err:
 	fputs("</tbody></table>", fp);
 
-	git_object_free(obj);
 	git_commit_free(commit);
 	git_tree_free(tree);
 
@@ -778,8 +783,7 @@ writerefs(FILE *fp)
 	git_reference *dref = NULL, *r, *ref = NULL;
 	git_reference_iterator *it = NULL;
 	git_reference **refs = NULL;
-	size_t count, i, j, len, refcount = 0;
-	const char *cols[] = { "Branch", "Tag" }; /* first column title */
+	size_t count, i, j, refcount = 0;
 	const char *titles[] = { "Branches", "Tags" };
 	const char *ids[] = { "branches", "tags" };
 	const char *name;
@@ -826,11 +830,9 @@ writerefs(FILE *fp)
 
 			/* print header if it has an entry (first). */
 			if (++count == 1) {
-				fprintf(fp, "<h2>%s</h2><table id=\"%s\"><thead>\n<tr><td>%s</td>"
-				      "<td>Age</td><td>Commit message</td>"
-				      "<td>Author</td><td>Files</td><td class=\"num\">+</td>"
-				      "<td class=\"num\">-</td></tr>\n</thead><tbody>\n",
-				      titles[j], ids[j], cols[j]);
+				fprintf(fp, "<h2>%s</h2><table id=\"%s\"><thead>\n<tr><td>Name</td>"
+				      "<td>Last commit date</td><td>Author</td>\n</tr>\n</thead><tbody>\n",
+				      titles[j], ids[j]);
 			}
 
 			relpath = "";
@@ -842,28 +844,8 @@ writerefs(FILE *fp)
 			if (ci->author)
 				printtimeshort(fp, &(ci->author->when));
 			fputs("</td><td>", fp);
-			if (ci->summary) {
-				if (j)
-					fprintf(fp, "<a href=\"%scommit/%s.html\">",
-					        relpath, ci->oid);
-				if ((len = strlen(ci->summary)) > summarylen) {
-					xmlencode(fp, ci->summary, summarylen - 1);
-					fputs("…", fp);
-				} else {
-					xmlencode(fp, ci->summary, len);
-				}
-				if (j)
-					fputs("</a>", fp);
-			}
-			fputs("</td><td>", fp);
 			if (ci->author)
 				xmlencode(fp, ci->author->name, strlen(ci->author->name));
-			fputs("</td><td class=\"num\">", fp);
-			fprintf(fp, "%zu", ci->filecount);
-			fputs("</td><td class=\"num\">", fp);
-			fprintf(fp, "+%zu", ci->addcount);
-			fputs("</td><td class=\"num\">", fp);
-			fprintf(fp, "-%zu", ci->delcount);
 			fputs("</td></tr>\n", fp);
 
 			relpath = "../";
@@ -876,7 +858,7 @@ writerefs(FILE *fp)
 		}
 		/* table footer */
 		if (count)
-			fputs("</tbody></table>", fp);
+			fputs("</tbody></table><br/>", fp);
 	}
 
 err:
@@ -894,10 +876,11 @@ int
 main(int argc, char *argv[])
 {
 	git_object *obj = NULL;
+	const git_oid *head = NULL;
 	const git_error *e = NULL;
 	FILE *fp, *fpread;
 	char path[PATH_MAX], *p;
-	int status;
+	int r, status;
 
 	if (argc != 2) {
 		fprintf(stderr, "%s <repodir>\n", argv[0]);
@@ -914,6 +897,12 @@ main(int argc, char *argv[])
 		return status;
 	}
 
+	/* find HEAD */
+	if (git_revparse_single(&obj, repo, "HEAD"))
+		return 1;
+	head = git_object_id(obj);
+	git_object_free(obj);
+
 	/* use directory name as name */
 	name = xbasename(repodir);
 
@@ -925,11 +914,17 @@ main(int argc, char *argv[])
 			*p = '\0';
 
 	/* read description or .git/description */
-	snprintf(path, sizeof(path), "%s%s%s",
+	r = snprintf(path, sizeof(path), "%s%s%s",
 		repodir, repodir[strlen(repodir)] == '/' ? "" : "/", "description");
+	if (r == -1 || (size_t)r >= sizeof(path))
+		errx(1, "path truncated: '%s%s%s'",
+	                repodir, repodir[strlen(repodir)] == '/' ? "" : "/", "description");
 	if (!(fpread = fopen(path, "r"))) {
-		snprintf(path, sizeof(path), "%s%s%s",
+		r = snprintf(path, sizeof(path), "%s%s%s",
 			repodir, repodir[strlen(repodir)] == '/' ? "" : "/", ".git/description");
+		if (r == -1 || (size_t)r >= sizeof(path))
+			errx(1, "path truncated: '%s%s%s'",
+		                repodir, repodir[strlen(repodir)] == '/' ? "" : "/", ".git/description");
 		fpread = fopen(path, "r");
 	}
 	if (fpread) {
@@ -939,11 +934,17 @@ main(int argc, char *argv[])
 	}
 
 	/* read url or .git/url */
-	snprintf(path, sizeof(path), "%s%s%s",
+	r = snprintf(path, sizeof(path), "%s%s%s",
 		repodir, repodir[strlen(repodir)] == '/' ? "" : "/", "url");
+	if (r == -1 || (size_t)r >= sizeof(path))
+		errx(1, "path truncated: '%s%s%s'",
+		        repodir, repodir[strlen(repodir)] == '/' ? "" : "/", "url");
 	if (!(fpread = fopen(path, "r"))) {
-		snprintf(path, sizeof(path), "%s%s%s",
+		r = snprintf(path, sizeof(path), "%s%s%s",
 			repodir, repodir[strlen(repodir)] == '/' ? "" : "/", ".git/url");
+		if (r == -1 || (size_t)r >= sizeof(path))
+			errx(1, "path truncated: '%s%s%s'",
+			        repodir, repodir[strlen(repodir)] == '/' ? "" : "/", ".git/url");
 		fpread = fopen(path, "r");
 	}
 	if (fpread) {
@@ -963,21 +964,22 @@ main(int argc, char *argv[])
 	/* log for HEAD */
 	fp = efopen("log.html", "w");
 	relpath = "";
-	writeheader(fp);
-	writelog(fp, "HEAD");
+	writeheader(fp, "Log");
+	mkdir("commit", 0755);
+	writelog(fp, head);
 	writefooter(fp);
 	fclose(fp);
 
 	/* files for HEAD */
 	fp = efopen("files.html", "w");
-	writeheader(fp);
-	writefiles(fp, "HEAD");
+	writeheader(fp, "Files");
+	writefiles(fp, head, "HEAD");
 	writefooter(fp);
 	fclose(fp);
 
 	/* summary page with branches and tags */
 	fp = efopen("refs.html", "w");
-	writeheader(fp);
+	writeheader(fp, "Refs");
 	writerefs(fp);
 	writefooter(fp);
 	fclose(fp);
